"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.LoginController = void 0;
const tslib_1 = require("tslib");
// Copyright (c) 2022 Sourcefuse Technologies
//
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT
const context_1 = require("@loopback/context");
const repository_1 = require("@loopback/repository");
const rest_1 = require("@loopback/rest");
const core_1 = require("@sourceloop/core");
const crypto_1 = require("crypto");
const jwt = tslib_1.__importStar(require("jsonwebtoken"));
const loopback4_authentication_1 = require("loopback4-authentication");
const loopback4_authorization_1 = require("loopback4-authorization");
const moment_timezone_1 = tslib_1.__importDefault(require("moment-timezone"));
const keys_1 = require("../../keys");
const providers_1 = require("../../providers");
const repositories_1 = require("../../repositories");
const tenant_config_repository_1 = require("../../repositories/tenant-config.repository");
const services_1 = require("../../services");
const _1 = require("./");
const auth_user_model_1 = require("./models/auth-user.model");
const reset_password_dto_1 = require("./models/reset-password.dto");
const token_response_dto_1 = require("./models/token-response.dto");
let LoginController = class LoginController {
    constructor(client, user, authClientRepository, userRepo, otpCacheRepo, roleRepo, utPermsRepo, userResourcesRepository, userTenantRepo, refreshTokenRepo, revokedTokensRepo, tenantConfigRepo, userCredsRepository, req, logger, getJwtPayload, loginHelperService, getAuthCode) {
        this.client = client;
        this.user = user;
        this.authClientRepository = authClientRepository;
        this.userRepo = userRepo;
        this.otpCacheRepo = otpCacheRepo;
        this.roleRepo = roleRepo;
        this.utPermsRepo = utPermsRepo;
        this.userResourcesRepository = userResourcesRepository;
        this.userTenantRepo = userTenantRepo;
        this.refreshTokenRepo = refreshTokenRepo;
        this.revokedTokensRepo = revokedTokensRepo;
        this.tenantConfigRepo = tenantConfigRepo;
        this.userCredsRepository = userCredsRepository;
        this.req = req;
        this.logger = logger;
        this.getJwtPayload = getJwtPayload;
        this.loginHelperService = loginHelperService;
        this.getAuthCode = getAuthCode;
    }
    async login(req, client, user) {
        await this.loginHelperService.verifyClientUserLogin(req, client, user);
        try {
            if (!this.user || !this.client) {
                // Control should never reach here
                this.logger.error(`${"Client Invalid" /* AuthErrorKeys.ClientInvalid */} :: Control should never reach here`);
                throw new rest_1.HttpErrors.Unauthorized("Client Invalid" /* AuthErrorKeys.ClientInvalid */);
            }
            const token = await this.getAuthCode(this.client, this.user);
            return {
                code: token,
            };
        }
        catch (error) {
            this.logger.error(error);
            throw new rest_1.HttpErrors.Unauthorized("Invalid Credentials" /* AuthErrorKeys.InvalidCredentials */);
        }
    }
    async loginWithClientUser(req) {
        var _a;
        await this.loginHelperService.verifyClientUserLogin(req, this.client, this.user);
        try {
            if (!this.user || !this.client) {
                // Control should never reach here
                this.logger.error(`${"Client Invalid" /* AuthErrorKeys.ClientInvalid */} :: Control should never reach here`);
                throw new rest_1.HttpErrors.Unauthorized("Client Invalid" /* AuthErrorKeys.ClientInvalid */);
            }
            const payload = {
                clientId: this.client.clientId,
                user: this.user,
            };
            if (((_a = payload.user) === null || _a === void 0 ? void 0 : _a.id) &&
                !(await this.userRepo.firstTimeUser(payload.user.id))) {
                await this.userRepo.updateLastLogin(payload.user.id);
            }
            return await this.createJWT(payload, this.client);
        }
        catch (error) {
            this.logger.error(error);
            throw new rest_1.HttpErrors.Unauthorized("Invalid Credentials" /* AuthErrorKeys.InvalidCredentials */);
        }
    }
    async getToken(req, codeReader) {
        const authClient = await this.authClientRepository.findOne({
            where: {
                clientId: req.clientId,
            },
        });
        if (!authClient) {
            throw new rest_1.HttpErrors.Unauthorized("Client Invalid" /* AuthErrorKeys.ClientInvalid */);
        }
        try {
            const code = await codeReader(req.code);
            const payload = jwt.verify(code, authClient.secret, {
                audience: req.clientId,
                issuer: process.env.JWT_ISSUER,
                algorithms: ['HS256'],
            });
            if (payload.mfa) {
                throw new rest_1.HttpErrors.Unauthorized("User Verification Failed" /* AuthErrorKeys.UserVerificationFailed */);
            }
            if (payload.userId &&
                !(await this.userRepo.firstTimeUser(payload.userId))) {
                await this.userRepo.updateLastLogin(payload.userId);
            }
            return await this.createJWT(payload, authClient);
        }
        catch (error) {
            this.logger.error(error);
            if (error.name === 'TokenExpiredError') {
                throw new rest_1.HttpErrors.Unauthorized("Code Expired" /* AuthErrorKeys.CodeExpired */);
            }
            else if (rest_1.HttpErrors.HttpError.prototype.isPrototypeOf(error)) {
                throw error;
            }
            else {
                throw new rest_1.HttpErrors.Unauthorized("Invalid Credentials" /* AuthErrorKeys.InvalidCredentials */);
            }
        }
    }
    async exchangeToken(req, deviceId, token) {
        const refreshPayload = await this.refreshTokenRepo.get(req.refreshToken);
        if (!refreshPayload) {
            throw new rest_1.HttpErrors.Unauthorized("Token Expired" /* AuthErrorKeys.TokenExpired */);
        }
        const authClient = await this.authClientRepository.findOne({
            where: {
                clientId: refreshPayload.clientId,
            },
        });
        if (!authClient) {
            throw new rest_1.HttpErrors.Unauthorized("Client Invalid" /* AuthErrorKeys.ClientInvalid */);
        }
        const accessToken = token === null || token === void 0 ? void 0 : token.split(' ')[1];
        if (!accessToken || refreshPayload.accessToken !== accessToken) {
            throw new rest_1.HttpErrors.Unauthorized("Token Invalid" /* AuthErrorKeys.TokenInvalid */);
        }
        await this.revokedTokensRepo.set(refreshPayload.accessToken, {
            token: refreshPayload.accessToken,
        });
        await this.refreshTokenRepo.delete(req.refreshToken);
        return this.createJWT({
            clientId: refreshPayload.clientId,
            userId: refreshPayload.userId,
            externalAuthToken: refreshPayload.externalAuthToken,
            externalRefreshToken: refreshPayload.externalRefreshToken,
        }, authClient);
    }
    async resetPassword(req, auth, currentUser) {
        const token = auth === null || auth === void 0 ? void 0 : auth.replace(/bearer /i, '');
        if (!token || !req.refreshToken) {
            throw new rest_1.HttpErrors.UnprocessableEntity("TokenMissing" /* AuthenticateErrorKeys.TokenMissing */);
        }
        const refreshTokenModel = await this.refreshTokenRepo.get(req.refreshToken);
        if (refreshTokenModel.accessToken !== token) {
            throw new rest_1.HttpErrors.Unauthorized("Token Invalid" /* AuthErrorKeys.TokenInvalid */);
        }
        if (refreshTokenModel.username !== req.username ||
            currentUser.username !== req.username) {
            throw new rest_1.HttpErrors.Forbidden("NotAllowedAccess" /* AuthorizeErrorKeys.NotAllowedAccess */);
        }
        if (req.password && req.password.length <= 0) {
            throw new rest_1.HttpErrors.BadRequest("PasswordInvalid" /* AuthenticateErrorKeys.PasswordInvalid */);
        }
        let changePasswordResponse;
        if (req.oldPassword) {
            changePasswordResponse = await this.userRepo.updatePassword(req.username, req.oldPassword, req.password);
        }
        else {
            changePasswordResponse = await this.userRepo.changePassword(req.username, req.password);
        }
        if (!changePasswordResponse) {
            throw new rest_1.HttpErrors.UnprocessableEntity('Unable to set password !');
        }
        const userTenant = await this.userTenantRepo.findOne({
            where: {
                userId: changePasswordResponse.id,
                tenantId: currentUser.tenantId,
            },
        });
        if (!userTenant) {
            throw new rest_1.HttpErrors.Unauthorized("UserInactive" /* AuthenticateErrorKeys.UserInactive */);
        }
        if (userTenant.status && userTenant.status < 1 /* UserStatus.ACTIVE */) {
            await this.userRepo.userTenants(changePasswordResponse.id).patch({
                status: 1 /* UserStatus.ACTIVE */,
            });
        }
        await this.revokedTokensRepo.set(token, { token });
        await this.refreshTokenRepo.delete(req.refreshToken);
        return new core_1.SuccessResponse({
            success: true,
        });
    }
    async me() {
        if (!this.user) {
            throw new rest_1.HttpErrors.Unauthorized("Token Invalid" /* AuthErrorKeys.TokenInvalid */);
        }
        delete this.user.deviceInfo;
        return new auth_user_model_1.AuthUser(this.user);
    }
    async createJWT(payload, authClient) {
        try {
            const size = 32;
            const ms = 1000;
            let user;
            if (payload.user) {
                user = payload.user;
            }
            else if (payload.userId) {
                user = await this.userRepo.findById(payload.userId, {
                    include: [
                        {
                            relation: 'defaultTenant',
                        },
                    ],
                });
                if (payload.externalAuthToken && payload.externalRefreshToken) {
                    user.externalAuthToken = payload.externalAuthToken;
                    user.externalRefreshToken =
                        payload.externalRefreshToken;
                }
            }
            else {
                // Do nothing and move ahead
            }
            if (!user) {
                throw new rest_1.HttpErrors.Unauthorized("UserDoesNotExist" /* AuthenticateErrorKeys.UserDoesNotExist */);
            }
            const data = await this.getJwtPayload(user, authClient);
            const accessToken = jwt.sign(data, process.env.JWT_SECRET, {
                expiresIn: authClient.accessTokenExpiration,
                issuer: process.env.JWT_ISSUER,
                algorithm: 'HS256',
            });
            const refreshToken = (0, crypto_1.randomBytes)(size).toString('hex');
            // Set refresh token into redis for later verification
            await this.refreshTokenRepo.set(refreshToken, {
                clientId: authClient.clientId,
                userId: user.id,
                username: user.username,
                accessToken,
                externalAuthToken: user.externalAuthToken,
                externalRefreshToken: user.externalRefreshToken,
            }, { ttl: authClient.refreshTokenExpiration * ms });
            return new token_response_dto_1.TokenResponse({
                accessToken,
                refreshToken,
                expires: (0, moment_timezone_1.default)()
                    .add(authClient.accessTokenExpiration, 's')
                    .toDate()
                    .getTime(),
            });
        }
        catch (error) {
            this.logger.error(error);
            if (rest_1.HttpErrors.HttpError.prototype.isPrototypeOf(error)) {
                throw error;
            }
            else {
                throw new rest_1.HttpErrors.Unauthorized("Invalid Credentials" /* AuthErrorKeys.InvalidCredentials */);
            }
        }
    }
};
tslib_1.__decorate([
    (0, loopback4_authentication_1.authenticateClient)("client-password" /* STRATEGY.CLIENT_PASSWORD */),
    (0, loopback4_authentication_1.authenticate)("local" /* STRATEGY.LOCAL */),
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.post)('/auth/login', {
        description: 'Gets you the code that will be used for getting token (webapps)',
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'Auth Code that you can use to generate access and refresh tokens using the POST /auth/token API',
                content: {
                    [core_1.CONTENT_TYPE.JSON]: Object,
                },
            },
            ...core_1.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__param(1, (0, context_1.inject)(loopback4_authentication_1.AuthenticationBindings.CURRENT_CLIENT)),
    tslib_1.__param(2, (0, context_1.inject)(loopback4_authentication_1.AuthenticationBindings.CURRENT_USER)),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [_1.LoginRequest, Object, Object]),
    tslib_1.__metadata("design:returntype", Promise)
], LoginController.prototype, "login", null);
tslib_1.__decorate([
    (0, loopback4_authentication_1.authenticateClient)("client-password" /* STRATEGY.CLIENT_PASSWORD */),
    (0, loopback4_authentication_1.authenticate)("OAuth2 resource owner grant" /* STRATEGY.OAUTH2_RESOURCE_OWNER_GRANT */),
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.post)('/auth/login-token', {
        description: 'Gets you refresh token and access token in one hit. (mobile app)',
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'Token Response Model',
                content: {
                    [core_1.CONTENT_TYPE.JSON]: {
                        schema: { [core_1.X_TS_TYPE]: token_response_dto_1.TokenResponse },
                    },
                },
            },
            ...core_1.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [_1.LoginRequest]),
    tslib_1.__metadata("design:returntype", Promise)
], LoginController.prototype, "loginWithClientUser", null);
tslib_1.__decorate([
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.post)('/auth/token', {
        description: 'Send the code received from the POST /auth/login api and get refresh token and access token (webapps)',
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'Token Response',
                content: {
                    [core_1.CONTENT_TYPE.JSON]: {
                        schema: { [core_1.X_TS_TYPE]: token_response_dto_1.TokenResponse },
                    },
                },
            },
            ...core_1.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__param(1, (0, context_1.inject)(providers_1.AuthCodeBindings.CODEREADER_PROVIDER)),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [_1.AuthTokenRequest, Function]),
    tslib_1.__metadata("design:returntype", Promise)
], LoginController.prototype, "getToken", null);
tslib_1.__decorate([
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.post)('/auth/token-refresh', {
        security: core_1.OPERATION_SECURITY_SPEC,
        description: 'Gets you a new access and refresh token once your access token is expired. (both mobile and web)\n',
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'New Token Response',
                content: {
                    [core_1.CONTENT_TYPE.JSON]: {
                        schema: { [core_1.X_TS_TYPE]: token_response_dto_1.TokenResponse },
                    },
                },
            },
            ...core_1.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__param(1, rest_1.param.header.string('device_id')),
    tslib_1.__param(2, rest_1.param.header.string('Authorization')),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [_1.AuthRefreshTokenRequest, String, String]),
    tslib_1.__metadata("design:returntype", Promise)
], LoginController.prototype, "exchangeToken", null);
tslib_1.__decorate([
    (0, loopback4_authentication_1.authenticate)("bearer" /* STRATEGY.BEARER */, { passReqToCallback: true }),
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.patch)(`auth/change-password`, {
        security: core_1.OPERATION_SECURITY_SPEC,
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'If User password successfully changed.',
            },
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)({
        content: {
            [core_1.CONTENT_TYPE.JSON]: {
                schema: (0, rest_1.getModelSchemaRef)(reset_password_dto_1.ResetPassword, { partial: true }),
            },
        },
    })),
    tslib_1.__param(1, rest_1.param.header.string('Authorization')),
    tslib_1.__param(2, (0, context_1.inject)(loopback4_authentication_1.AuthenticationBindings.CURRENT_USER)),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [reset_password_dto_1.ResetPassword, String, Object]),
    tslib_1.__metadata("design:returntype", Promise)
], LoginController.prototype, "resetPassword", null);
tslib_1.__decorate([
    (0, loopback4_authentication_1.authenticate)("bearer" /* STRATEGY.BEARER */, {
        passReqToCallback: true,
    }),
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.get)('/auth/me', {
        security: core_1.OPERATION_SECURITY_SPEC,
        description: 'To get the user details',
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'User Object',
                content: {
                    [core_1.CONTENT_TYPE.JSON]: auth_user_model_1.AuthUser,
                },
            },
            ...core_1.ErrorCodes,
        },
    }),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", []),
    tslib_1.__metadata("design:returntype", Promise)
], LoginController.prototype, "me", null);
LoginController = tslib_1.__decorate([
    tslib_1.__param(0, (0, context_1.inject)(loopback4_authentication_1.AuthenticationBindings.CURRENT_CLIENT)),
    tslib_1.__param(1, (0, context_1.inject)(loopback4_authentication_1.AuthenticationBindings.CURRENT_USER)),
    tslib_1.__param(2, (0, repository_1.repository)(repositories_1.AuthClientRepository)),
    tslib_1.__param(3, (0, repository_1.repository)(repositories_1.UserRepository)),
    tslib_1.__param(4, (0, repository_1.repository)(repositories_1.OtpCacheRepository)),
    tslib_1.__param(5, (0, repository_1.repository)(repositories_1.RoleRepository)),
    tslib_1.__param(6, (0, repository_1.repository)(repositories_1.UserLevelPermissionRepository)),
    tslib_1.__param(7, (0, repository_1.repository)(repositories_1.UserLevelResourceRepository)),
    tslib_1.__param(8, (0, repository_1.repository)(repositories_1.UserTenantRepository)),
    tslib_1.__param(9, (0, repository_1.repository)(repositories_1.RefreshTokenRepository)),
    tslib_1.__param(10, (0, repository_1.repository)(repositories_1.RevokedTokenRepository)),
    tslib_1.__param(11, (0, repository_1.repository)(tenant_config_repository_1.TenantConfigRepository)),
    tslib_1.__param(12, (0, repository_1.repository)(repositories_1.UserCredentialsRepository)),
    tslib_1.__param(13, (0, context_1.inject)(rest_1.RestBindings.Http.REQUEST)),
    tslib_1.__param(14, (0, context_1.inject)(core_1.LOGGER.LOGGER_INJECT)),
    tslib_1.__param(15, (0, context_1.inject)(keys_1.AuthServiceBindings.JWTPayloadProvider)),
    tslib_1.__param(16, (0, context_1.inject)('services.LoginHelperService')),
    tslib_1.__param(17, (0, context_1.inject)(providers_1.AuthCodeBindings.AUTH_CODE_GENERATOR_PROVIDER)),
    tslib_1.__metadata("design:paramtypes", [Object, Object, repositories_1.AuthClientRepository,
        repositories_1.UserRepository,
        repositories_1.OtpCacheRepository,
        repositories_1.RoleRepository,
        repositories_1.UserLevelPermissionRepository,
        repositories_1.UserLevelResourceRepository,
        repositories_1.UserTenantRepository,
        repositories_1.RefreshTokenRepository,
        repositories_1.RevokedTokenRepository,
        tenant_config_repository_1.TenantConfigRepository,
        repositories_1.UserCredentialsRepository, Object, Object, Function, services_1.LoginHelperService, Function])
], LoginController);
exports.LoginController = LoginController;
//# sourceMappingURL=login.controller.js.map